# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

from enum import Enum
import json
import logging
import re
from typing import Any, Dict, List, Optional, Tuple

import requests
import urllib3
import yaml

from .apic import ApicObject
from .const import RN_PREFIX_CLASSNAME_MAPPINGS
from .nae import NAE
from .ndi import NDI

logger = logging.getLogger(__name__)


class PCV:
    class Platform(Enum):
        NDI = 1
        NAE = 2

    def __init__(
        self,
        hostname_ip: str,
        username: str,
        password: str,
        domain: str,
        timeout: int,
        platform: Platform,
    ):
        self.platform = platform
        if platform is self.Platform.NDI:
            self.ndi = NDI(hostname_ip, username, password, domain, timeout)
        elif platform is self.Platform.NAE:
            self.nae = NAE(hostname_ip, username, password, domain, timeout)
        self.root = ApicObject("root", {}, [], None)
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    def _resolve_tf_classnames(self, root: ApicObject, tf_plan: Any) -> None:
        """Helper function to resolve missing class names and key attributes using the Terraform plan"""
        if root.cl is None:
            dn = root["dn"]
            for change in tf_plan.get("resource_changes", []):
                section = (
                    "after" if change["change"].get("after") is not None else "before"
                )
                plan_dn = change["change"].get(section, {}).get("dn")
                if dn == plan_dn:
                    root.cl = change["change"].get(section, {}).get("class_name")
                    root.attributes["name"] = (
                        change["change"].get(section, {}).get("content", {}).get("name")
                    )

        for child in root.children:
            self._resolve_tf_classnames(child, tf_plan)

    def _resolve_static_classnames(self, root: ApicObject) -> Optional[str]:
        """Helper function to resolve missing class names and key attributes using static mappings"""
        if root.cl is None:
            parts = str(root["dn"]).split("/")[-1].split("-", 1)
            prefix = parts[0]
            name = parts[1] if len(parts) > 1 else None
            if prefix in RN_PREFIX_CLASSNAME_MAPPINGS:
                mapping = RN_PREFIX_CLASSNAME_MAPPINGS[prefix]
                root.cl = mapping.get("class")
                for key in mapping.get("keys", []):
                    key_attribute = key.get("attribute")
                    key_regex = key.get("regex")
                    if (
                        key_attribute is not None
                        and key_regex is not None
                        and name is not None
                    ):
                        regex = re.compile(key_regex)
                        mo = regex.search(name)
                        if mo is not None:
                            root.attributes[key_attribute] = mo.group()
            if root.cl is None:
                return str(root["dn"])
        for child in root.children:
            self._resolve_static_classnames(child)
        return None

    def _load_json_objects(
        self, json_dict: Dict[Any, Any], parent: Optional[ApicObject] = None
    ) -> Optional[ApicObject]:
        """Helper function to load JSON objects into object tree"""
        new_obj = None
        for k, v in json_dict.items():
            new_obj = ApicObject(k, v.get("attributes"), [], parent)
            if parent:
                parent.children.append(new_obj)
            for child in v.get("children", []):
                self._load_json_objects(child, new_obj)
        return new_obj

    def load_json_files(self, filenames: List[str]) -> Optional[str]:
        """Load objects from JSON files into object tree"""
        for filename in filenames:
            try:
                with open(filename, "r") as file:
                    inv = json.loads(file.read())
                    if "imdata" in inv:
                        for item in inv["imdata"]:
                            obj = self._load_json_objects(item)
                            self.root.insert(obj)
                    else:
                        obj = self._load_json_objects(inv)
                        self.root.insert(obj)
            except:  # noqa E722
                logger.error("Failed to load JSON file: {}".format(filename))
        err = self._resolve_static_classnames(self.root)
        if err is not None:
            return err
        return None

    def load_tf_plan(self, filename: str) -> Optional[str]:
        """Load changed objects from Terraform plan into object tree"""
        tf_plan = None
        with open(filename) as file:
            tf_plan = json.load(file)

        for change in tf_plan.get("resource_changes", []):
            if change.get("type") == "aci_rest_managed":
                action = change["change"].get("actions", [])
                if "create" in action or "update" in action or "delete" in action:
                    if "delete" in action:
                        classname = change["change"].get("before", {}).get("class_name")
                        attributes = {}
                        attributes = change["change"].get("before", {}).get("content")
                        attributes["status"] = "deleted"
                        attributes["dn"] = change["change"].get("before", {}).get("dn")
                    else:
                        classname = change["change"].get("after", {}).get("class_name")
                        attributes = change["change"].get("after", {}).get("content")
                        attributes["dn"] = change["change"].get("after", {}).get("dn")
                    attributes = {k: v for (k, v) in attributes.items() if v != ""}
                    obj = ApicObject(classname, attributes, [], None)
                    self.root.insert(obj)

        err = self._resolve_static_classnames(self.root)
        if err is not None:
            return err
        self._resolve_tf_classnames(self.root, tf_plan)
        return None

    def _write_pcv_events(self, events: List[Any], file: str) -> None:
        with open(file, "w") as fh:
            fh.write(yaml.dump(events, default_flow_style=False))

    def _write_pcv_url(self, url: str, file: str) -> None:
        with open(file, "w") as fh:
            fh.write(url)

    def ndi_pcv(
        self,
        name: str,
        group: str,
        site: str,
        suppress_events: str,
        file_summary: str,
        file_url: str,
    ) -> Tuple[Optional[requests.Response], Optional[List[Any]], Optional[str]]:
        """Trigger an NDI pre-change validation"""
        if not len(self.root.children):
            logger.info("No updates planned. No need to trigger a pre-change analysis.")
            return None, None, None
        logger.debug("Proposed change (JSON): {}".format(self.root[0]))
        err, job_id = self.ndi.start_pcv(name, group, site, str(self.root[0]))
        if err is not None:
            return err, None, None
        err, epoch_job_id = self.ndi.wait_pcv(group, site, str(job_id))
        if err is not None:
            return err, None, None
        err, events = self.ndi.get_pcv_results(
            group, site, str(epoch_job_id), suppress_events
        )
        if err is not None:
            return err, None, None
        err, url = self.ndi.get_pcv_url()
        if err is not None:
            return err, None, None
        if file_summary and events:
            self._write_pcv_events(events, file_summary)
        if file_url and url is not None:
            self._write_pcv_url(url, file_url)
        return None, events, url

    def nae_pcv(
        self,
        name: str,
        group: str,
        suppress_events: str,
        file_summary: str,
        file_url: str,
    ) -> Tuple[Optional[requests.Response], Optional[List[Any]], Optional[str]]:
        """Trigger an NAE pre-change validation"""
        if not len(self.root.children):
            logger.info("No updates planned. No need to trigger a pre-change analysis.")
            return None, None, None
        logger.debug("Proposed change (JSON): {}".format(self.root[0]))
        err, job_id = self.nae.start_pcv(name, group, str(self.root[0]))
        if err is not None:
            return err, None, None
        err, epoch_job_id = self.nae.wait_pcv(str(job_id))
        if err is not None:
            return err, None, None
        err, events = self.nae.get_pcv_results(str(epoch_job_id), suppress_events)
        if err is not None:
            return err, None, None
        err, url = self.nae.get_pcv_url(str(job_id))
        if err is not None:
            return err, None, None
        if file_summary and events:
            self._write_pcv_events(events, file_summary)
        if file_url and url is not None:
            self._write_pcv_url(url, file_url)
        return None, events, url



# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

try:
    from importlib.metadata import PackageNotFoundError, version  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError, version
try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

import json
import logging
from typing import Dict, List, Optional, Union

logger = logging.getLogger(__name__)


class ApicObject:
    def __init__(
        self,
        cl: Optional[str],
        attributes: Dict[str, str],
        children: List["ApicObject"],
        parent: Optional["ApicObject"],
    ):
        self.cl = cl
        self.attributes = attributes
        self.children = children
        self.parent = parent

    def update(
        self,
        attributes: Dict[str, str],
        children: List["ApicObject"],
    ) -> None:
        """Update object attributes, classname and children"""
        self.attributes.update(attributes)
        for child in children:
            dn = child.attributes.get("dn")
            name = child.attributes.get("name")
            found = False
            # look for existing object with dn
            if dn is not None:
                for c in self.children:
                    if c.attributes.get("dn") == dn:
                        if child.cl != c.cl:
                            continue
                        c.update(child.attributes, child.children)
                        found = True
            if found:
                continue
            # look for existing object with name
            if name is not None:
                for c in self.children:
                    if c.attributes.get("name") == name and child.cl == c.cl:
                        c.update(child.attributes, child.children)
                        found = True
            if found:
                continue
            # add as a new child
            self.children.append(
                ApicObject(child.cl, child.attributes, child.children, self)
            )

    def find(self, dn: str = "", cl: str = "") -> List["ApicObject"]:
        """Find objects by dn or classname in subtree"""
        result: List["ApicObject"] = []
        if not dn and not cl:
            return result
        elif not cl:
            if self.attributes.get("dn") == dn:
                result.append(self)
        elif not dn:
            if self.cl == cl:
                result.append(self)
        elif self.attributes.get("dn") == dn and self.cl == cl:
            result.append(self)
        for child in self.children:
            objs = child.find(dn=dn, cl=cl)
            result.extend(objs)
        return result

    def _index_of_last_dn_delimiter(self, dn: str) -> int:
        """Helper function to return index of last delimiter ('/') in DN string"""
        escaped = False
        index = len(dn) - 1
        for c in reversed(dn):
            if c == "]":
                escaped = True
            elif c == "[":
                escaped = False
            elif c == "/" and not escaped:
                return index
            index -= 1
        return -1

    def insert(self, obj: Optional["ApicObject"]) -> None:
        """Insert object in correct place in tree according to dn"""
        if obj is None:
            return
        dn = obj.attributes["dn"]
        o = self.find(dn=dn)
        if len(o) > 0:
            o[0].update(obj.attributes, obj.children)
        else:
            index = self._index_of_last_dn_delimiter(dn)
            if index == -1:
                self.children.append(obj)
                obj.parent = self
            else:
                parent_dn = dn[:index]
                o = self.find(dn=parent_dn)
                if len(o) > 0:
                    o[0].children.append(obj)
                    obj.parent = o[0]
                else:
                    new_obj = ApicObject(None, {"dn": parent_dn}, [obj], None)
                    obj.parent = new_obj
                    self.insert(new_obj)

    def add_child(
        self, cl: str, attributes: Dict[str, str], children: List["ApicObject"]
    ) -> "ApicObject":
        """Add child to object"""
        child = ApicObject(cl, attributes, children, self)
        self.children.append(child)
        return child

    def add_parent(self, cl: str, attributes: Dict[str, str]) -> "ApicObject":
        """Add parent to object"""
        if self.parent is not None:
            raise Exception(
                "ApicObject {} already has a parent.".format(str(ApicObject))
            )
        self.parent = ApicObject(cl, attributes, [self], None)
        return self.parent

    def get_root(self) -> Optional["ApicObject"]:
        """Get root object"""
        obj = self
        # max search depth 100
        for i in range(100):
            if obj.cl == "root":
                return obj
            elif obj.parent is not None:
                obj = obj.parent
            else:
                return None
        return None

    def __getitem__(self, key: Union[str, int]) -> Union["ApicObject", str, None]:
        """Get attribute if key is string otherwise child by index"""
        if isinstance(key, str):
            return self.attributes.get(key)
        else:
            return self.children[key]

    def __str__(self) -> str:
        """Return json string."""
        attr_string = ", ".join(
            ['"{}": {}'.format(k, json.dumps(v)) for k, v in self.attributes.items()]
        )
        child_string = ", ".join([str(c) for c in self.children])
        return '{{"{}": {{"attributes": {{{}}}, "children": [{}]}}}}'.format(
            self.cl, attr_string, child_string
        )



# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

from datetime import datetime
import json
import logging
import time
from typing import Any, List, Optional, Tuple

import requests
import yaml

logger = logging.getLogger(__name__)


class NDI:
    def __init__(
        self,
        hostname_ip: str,
        username: str,
        password: str,
        domain: str,
        timeout: int,
    ):
        self.hostname_ip = hostname_ip
        self.api_url = "https://{}/sedgeapi/v1/cisco-nir/api/api/telemetry/v2".format(
            hostname_ip
        )
        self.username = username
        self.password = password
        if domain == "Local":
            self.domain = "local"
        else:
            self.domain = domain
        self.timeout = timeout
        self.session = requests.Session()
        self.session.verify = False
        self.authenticated = False
        self.site_uuid = ""

    def _login(self) -> Optional[requests.Response]:
        """Helper function to authenticate and populate headers"""
        auth_payload = {
            "userName": self.username,
            "userPasswd": self.password,
            "domain": self.domain,
        }
        url = "https://{}/login".format(self.hostname_ip)
        resp = self.session.post(url, json=auth_payload)
        if resp.status_code != 200:
            logger.error("Login failed: {}".format(resp.json()))
            return resp
        self.authenticated = True
        return None

    def get_last_epoch_id(
        self, name: str, site: str
    ) -> Tuple[Optional[requests.Response], Optional[str]]:
        """Get last epoch ID of assurance group"""
        if not self.authenticated:
            err = self._login()
            if err is not None:
                return err, None

        url = "{}/events/insightsGroup/{}/fabric/{}/epochs?$size=1&$status=FINISHED&$epochType=ONLINE".format(
            self.api_url, name, site
        )
        resp = self.session.get(url)
        if resp.status_code != 200:
            logger.error("Get epoch id failed: {}".format(resp.json()))
            return resp, None

        try:
            epochs = json.loads(resp.content)["value"]["data"]
            epoch_id = epochs[0]["epochId"]
            self.site_uuid = epochs[0]["fabricId"]
            return None, epoch_id
        except KeyError:
            pass
        logger.error("Epoch ID could not be found: {}".format(resp.json()))
        return resp, None

    def start_pcv(
        self, name: str, group: str, site: str, json_data: str
    ) -> Tuple[Optional[requests.Response], Optional[str]]:
        """Start pre-change validation and return job ID"""
        if not self.authenticated:
            err = self._login()
            if err is not None:
                return err, None

        err, epoch_id = self.get_last_epoch_id(group, site)
        if err is not None:
            return err, None

        payload = {}
        payload["name"] = name
        payload["fabricUuid"] = self.site_uuid
        payload["baseEpochId"] = str(epoch_id)
        payload["allowUnsupportedObjectModification"] = "true"
        payload["uploadedFileName"] = "tmp.json"
        payload["assuranceEntityName"] = site

        files = [
            ("data", ("blob", json.dumps(payload), "application/json")),
            ("file", ("tmp.json", json_data, "application/json")),
        ]

        url = (
            "{}/config/insightsGroup/{}/fabric/{}/prechangeAnalysis/fileChanges".format(
                self.api_url, group, site
            )
        )
        resp = self.session.post(url, files=files)
        if resp.status_code != 200:
            logger.error("Start pre-change analysis failed: {}".format(resp.json()))
            return resp, None

        try:
            job_id = json.loads(resp.content)["value"]["data"]["jobId"]
            logger.info("Pre-change analysis started. Job ID: {}".format(job_id))
            return None, job_id
        except KeyError:
            pass
        logger.error("Job ID could not be found: {}".format(resp.json()))
        return resp, None

    def wait_pcv(
        self, group: str, site: str, job_id: str
    ) -> Tuple[Optional[requests.Response], Optional[str]]:
        """Wair for pre-change validation to complete and return epoch job ID"""
        if not self.authenticated:
            err = self._login()
            if err is not None:
                return err, None

        status = None
        start_time = datetime.now()
        while True:
            url = "{}/config/insightsGroup/{}/fabric/{}/prechangeAnalysis/{}".format(
                self.api_url, group, site, job_id
            )
            resp = self.session.get(url)
            if resp.status_code != 200:
                logger.error(
                    "Get pre-change analysis status failed: {}".format(resp.json())
                )
                return resp, None
            try:
                status = json.loads(resp.content)["value"]["data"]["analysisStatus"]
                if status == "COMPLETED":
                    break
            except KeyError:
                logger.error("Status could not be found: {}".format(resp.json()))
            delta_minutes = (datetime.now() - start_time).total_seconds() / 60
            if delta_minutes > self.timeout:
                break
            logger.info("Waiting for pre-change analysis to complete ...")
            time.sleep(10)

        try:
            epoch_job_id = json.loads(resp.content)["value"]["data"]["epochDeltaJobId"]
            logger.info(
                "Pre-change analysis completed. Epoch job ID: {}".format(epoch_job_id)
            )
            return None, epoch_job_id
        except KeyError:
            pass
        logger.error("Epoch job ID could not be found: {}".format(resp.json()))
        return resp, None

    def get_pcv_results(
        self, group: str, site: str, epoch_job_id: str, suppress_events: str
    ) -> Tuple[Optional[requests.Response], Optional[List[Any]]]:
        """Retrieve pre-change validation results"""
        if not self.authenticated:
            err = self._login()
            if err is not None:
                return err, None

        suppress_events_list = suppress_events.split(",")

        url = "{}/epochDelta/insightsGroup/{}/fabric/{}/job/{}/health/view/aggregateTable?category=ADC,CHANGE_ANALYSIS,TENANT_ENDPOINT,TENANT_FORWARDING,TENANT_SECURITY,RESOURCE_UTILIZATION,SYSTEM,COMPLIANCE&epochStatus=EPOCH2_ONLY&severity=EVENT_SEVERITY_CRITICAL,EVENT_SEVERITY_MAJOR,EVENT_SEVERITY_MINOR,EVENT_SEVERITY_WARNING,EVENT_SEVERITY_INFO".format(
            self.api_url, group, site, epoch_job_id
        )
        resp = self.session.get(url)
        if resp.status_code != 200:
            logger.error("Get PCV results failed: {}".format(resp.json()))
            return resp, None

        event_list = []
        try:
            for event in json.loads(resp.content)["entries"]:
                if int(event["count"]) > 0:
                    if (
                        str(event.get("severity")) == "info"
                        or str(event.get("mnemonicTitle")) in suppress_events_list
                    ):
                        continue
                    event_list.append(
                        {
                            "Category": event.get("category").title(),
                            "Count": event.get("count"),
                            "Description": event.get("anomalyStr"),
                            "Severity": event.get("severity"),
                        }
                    )
        except KeyError:
            logger.error("Could not find events: {}".format(resp.json()))
            return resp, None
        if event_list:
            logger.error(
                "The following anomalies have been raised:\n{}".format(
                    yaml.dump(event_list)
                )
            )
        return None, event_list

    def get_pcv_url(self) -> Tuple[Optional[requests.Response], Optional[str]]:
        """Get URL pointing to pre-change validation results"""
        if not self.authenticated:
            err = self._login()
            if err is not None:
                return err, None

        url = "https://{}/appcenter/cisco/nexus-insights/ui/#/changeManagement/preChangeAnalysis".format(
            self.hostname_ip
        )

        return None, url



# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

import logging
import sys
from typing import List

import click
import errorhandler

import nexus_pcv
from nexus_pcv.pcv import PCV

from . import options

logger = logging.getLogger(__name__)

error_handler = errorhandler.ErrorHandler()


def configure_logging(level: str) -> None:
    if level == "DEBUG":
        lev = logging.DEBUG
    elif level == "INFO":
        lev = logging.INFO
    elif level == "WARNING":
        lev = logging.WARNING
    elif level == "ERROR":
        lev = logging.ERROR
    else:
        lev = logging.CRITICAL
    logger = logging.getLogger()
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
    logger.addHandler(handler)
    logger.setLevel(lev)


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(nexus_pcv.__version__)
@click.option(
    "-v",
    "--verbosity",
    metavar="LVL",
    is_eager=True,
    type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]),
    help="Either CRITICAL, ERROR, WARNING, INFO or DEBUG.",
    default="WARNING",
)
@options.hostname_ip
@options.username
@options.password
@options.domain
@options.group
@options.site
@options.name
@options.suppress_events
@options.timeout
@options.file
@options.nac_tf_plan
@options.output_summary
@options.output_url
def main(
    verbosity: str,
    hostname_ip: str,
    username: str,
    password: str,
    domain: str,
    group: str,
    site: str,
    name: str,
    suppress_events: str,
    timeout: int,
    file: List[str],
    nac_tf_plan: str,
    output_summary: str,
    output_url: str,
) -> None:
    """A CLI tool to perform a pre-change validation on Nexus Dashboard Insights or Network Assurance Engine."""
    configure_logging(verbosity)

    platform = PCV.Platform.NDI if site else PCV.Platform.NAE
    pcv = PCV(hostname_ip, username, password, domain, timeout, platform)
    if file:
        pcv.load_json_files(file)
    if nac_tf_plan:
        pcv.load_tf_plan(nac_tf_plan)
    if platform is PCV.Platform.NDI:
        pcv.ndi_pcv(name, group, site, suppress_events, output_summary, output_url)
    else:
        pcv.nae_pcv(name, group, suppress_events, output_summary, output_url)
    exit()


def exit() -> None:
    if error_handler.fired:
        sys.exit(1)
    else:
        sys.exit(0)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nexus_pcv', 'nexus_pcv.cli']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'errorhandler>=2.0.1,<3.0.0',
 'pyyaml>=6.0,<7.0',
 'requests>=2.28.1,<3.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=2.0.0,<3.0.0']}

entry_points = \
{'console_scripts': ['nexus-pcv = nexus_pcv.cli.main:main']}

setup_kwargs = {
    'name': 'nexus-pcv',
    'version': '0.1.2',
    'description': 'A CLI tool to perform a pre-change validation on Nexus Dashboard Insights or Network Assurance Engine.',
    'long_description': '[![Tests](https://github.com/netascode/nexus-pcv/actions/workflows/test.yml/badge.svg)](https://github.com/netascode/nexus-pcv/actions/workflows/test.yml)\n![Python Support](https://img.shields.io/badge/python-3.7%20%7C%203.8%20%7C%203.9%20%7C%203.10-informational "Python Support: 3.7, 3.8, 3.9, 3.10")\n\n# nexus-pcv\n\nA CLI tool to perform a pre-change validation on Nexus Dashboard Insights or Network Assurance Engine. It can either work with provided JSON file(s) or a `terraform plan` output from a [Nexus as Code](https://cisco.com/go/nexusascode) project. It waits for the analysis to complete and evaluates the results.\n\n```\n$ nexus-pcv -h\nUsage: nexus-pcv [OPTIONS]\n\n  A CLI tool to perform a pre-change validation on Nexus Dashboard Insights or\n  Network Assurance Engine.\n\nOptions:\n  --version                   Show the version and exit.\n  -v, --verbosity LVL         Either CRITICAL, ERROR, WARNING, INFO or DEBUG.\n  -i, --hostname-ip TEXT      NAE/ND hostname or IP (required, env:\n                              PCV_HOSTNAME_IP).\n  -u, --username TEXT         NAE/ND username (required, env: PCV_USERNAME).\n  -p, --password TEXT         NAE/ND password (required, env: PCV_PASSWORD).\n  -d, --domain TEXT           NAE/ND login domain (optional, default: \'Local\',\n                              env: PCV_DOMAIN).\n  -g, --group TEXT            NAE assurance group name or NDI insights group\n                              name (required, env: PCV_GROUP).\n  -s, --site TEXT             NDI site or fabric name (optional, only required\n                              for NDI, env: PCV_SITE).\n  -n, --name TEXT             NAE/NDI pre-change validation name (optional,\n                              env: PCV_NAME).\n  -s, --suppress-events TEXT  NAE/NDI comma-separated list of events to\n                              suppress (optional, default: \'APP_EPG_NOT_DEPLOY\n                              ED,APP_EPG_HAS_NO_CONTRACT_IN_ENFORCED_VRF\',\n                              env: PCV_SUPPRESS_EVENTS).\n  -t, --timeout INTEGER       NAE/NDI pre-change validation timeout in minutes\n                              (optional, default: 15, env: PCV_TIMEOUT).\n  -f, --file FILE             NAE/NDI proposed change JSON file (optional,\n                              env: PCV_FILE).\n  -t, --nac-tf-plan FILE      NAE/NDI proposed change Terraform plan output\n                              (optional, env: PCV_NAC_TF_PLAN).\n  -o, --output-summary FILE   NAE/NDI summary of new events/anomalies written\n                              to a file (optional, env: PCV_OUTPUT_SUMMARY).\n  -r, --output-url FILE       NAE/NDI link (URL) to pre-change validation\n                              results written to a file (optional, env:\n                              PCV_OUTPUT_URL).\n  -h, --help                  Show this message and exit.\n```\n\n## Installation\n\nPython 3.7+ is required to install `nexus-pcv`. Don\'t have Python 3.7 or later? See [Python 3 Installation & Setup Guide](https://realpython.com/installing-python/).\n\n`nexus-pcv` can be installed in a virtual environment using `pip`:\n\n```\npip install nexus-pcv\n```\n\n## CI/CD Integration\n\nThe tool can easily be integrated with CI/CD workflows. Arguments can either be provided via command line or environment variables. The tool will exit with a non-zero exit code in case of an error or non-suppressed events being discovered during the pre-change analysis. The `--output-summary` and `--output-url` arguments can be used to write a summary and/or a link (URL) to a file, which can then be embedded into notifications (e.g., Webex).\n\n## *Nexus as Code* Integration\n\n*Nexus as Code* allows users to instantiate network fabrics in minutes using an easy to use, opinionated data model. More information about *Nexus as Code* can be found [here](https://cisco.com/go/nexusascode). A planned change can be validated before applying it to a production environment by running a `terraform plan` operation first and then providing the output to `nexus-pcv` to trigger a pre-change validation.\n\n```\nexport PCV_HOSTNAME_IP=10.1.1.1\nexport PCV_USERNAME=admin\nexport PCV_PASSWORD=Cisco123\nexport PCV_GROUP=LAB\nexport PCV_SITE=LAB1\nterraform plan -out=plan.tfplan\nterraform show -json plan.tfplan > plan.json\nnexus-pcv --name "PCV1" --nac-tf-plan plan.json\n```\n',
    'author': 'Daniel Schmidt',
    'author_email': 'danischm@cisco.com',
    'maintainer': 'Daniel Schmidt',
    'maintainer_email': 'danischm@cisco.com',
    'url': 'https://github.com/netascode/nexus-pcv',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

from datetime import datetime
import json
import logging
import time
from typing import Any, List, Optional, Tuple

import requests
import yaml

logger = logging.getLogger(__name__)


class NAE:
    def __init__(
        self,
        hostname_ip: str,
        username: str,
        password: str,
        domain: str,
        timeout: int,
    ):
        self.hostname_ip = hostname_ip
        self.api_url = "https://{}/nae/api/v1".format(hostname_ip)
        self.username = username
        self.password = password
        self.domain = domain
        self.timeout = timeout
        self.session = requests.Session()
        self.session.verify = False
        self.authenticated = False
        self.assurance_group_uuid = ""

    def _login(self) -> Optional[requests.Response]:
        """Helper function to authenticate and populate headers"""
        url = "{}/whoami".format(self.api_url)
        resp = self.session.get(url)
        if resp.status_code != 200:
            logger.error("Whoami failed: {}".format(resp.json()))
            return resp
        if "X-NAE-LOGIN-OTP" not in resp.headers:
            logger.error(
                "X-NAE-LOGIN-OTP header missing in whoami response: {}".format(
                    resp.json()
                )
            )
            return resp

        headers = {}
        headers["X-NAE-LOGIN-OTP"] = resp.headers.get("X-NAE-LOGIN-OTP")
        headers["Cookie"] = resp.headers.get("Set-Cookie")
        auth_payload = {
            "username": self.username,
            "password": self.password,
            "domain": self.domain,
        }
        url = "{}/login".format(self.api_url)
        resp = self.session.post(url, headers=headers, json=auth_payload)
        if resp.status_code != 200:
            logger.error("Login failed: {}".format(resp.json()))
            return resp
        if "X-NAE-CSRF-TOKEN" not in resp.headers:
            logger.error(
                "X-NAE-CSRF-TOKEN header missing in login response: {}".format(
                    resp.json()
                )
            )
            return resp

        self.session.headers["X-NAE-CSRF-TOKEN"] = str(
            resp.headers.get("X-NAE-CSRF-TOKEN")
        )
        self.session.headers["Cookie"] = str(resp.headers.get("Set-Cookie"))
        self.authenticated = True
        return None

    def get_assurance_group_uuid(
        self, name: str
    ) -> Tuple[Optional[requests.Response], Optional[str]]:
        """Get assurance group ID by its name"""
        if not self.authenticated:
            err = self._login()
            if err is not None:
                return err, None

        url = "{}/config-services/assured-networks/aci-fabric/".format(self.api_url)
        resp = self.session.get(url)
        if resp.status_code != 200:
            logger.error("Get assurance group failed: {}".format(resp.json()))
            return resp, None

        try:
            ags = json.loads(resp.content)["value"]["data"]
            for ag in ags:
                if ag["unique_name"] == name:
                    self.assurance_group_uuid = ag.get("uuid")
                    return None, ag.get("uuid")
        except KeyError:
            pass
        logger.error("UUID could not be found: {}".format(resp.json()))
        return resp, None

    def get_last_epoch_id(
        self, name: str
    ) -> Tuple[Optional[requests.Response], Optional[str]]:
        """Get last epoch ID of assurance group"""
        if not self.authenticated:
            err = self._login()
            if err is not None:
                return err, None

        err, uuid = self.get_assurance_group_uuid(name)
        if err is not None:
            return err, None

        url = "{}/event-services/assured-networks/{}/epochs?$sort=-collectionTimestamp".format(
            self.api_url, uuid
        )
        resp = self.session.get(url)
        if resp.status_code != 200:
            logger.error("Get epoch id failed: {}".format(resp.json()))
            return resp, None

        try:
            epochs = json.loads(resp.content)["value"]["data"]
            epoch_id = epochs[0]["epoch_id"]
            return None, epoch_id
        except KeyError:
            pass
        logger.error("Epoch ID could not be found: {}".format(resp.json()))
        return resp, None

    def start_pcv(
        self, name: str, ag_name: str, json_data: str
    ) -> Tuple[Optional[requests.Response], Optional[str]]:
        """Start pre-change validation and return job ID"""
        if not self.authenticated:
            err = self._login()
            if err is not None:
                return err, None

        err, epoch_id = self.get_last_epoch_id(ag_name)
        if err is not None:
            return err, None

        payload = {}
        payload["name"] = name
        payload["fabric_uuid"] = self.assurance_group_uuid
        payload["base_epoch_id"] = str(epoch_id)
        payload["stop_analysis"] = str(True)
        payload["allow_unsupported_object_modification"] = "true"
        payload["uploaded_file_name"] = "tmp.json"

        files = [
            ("data", ("blob", json.dumps(payload), "application/json")),
            ("file", ("tmp.json", json_data, "application/json")),
        ]

        url = "{}/config-services/prechange-analysis/file-changes".format(self.api_url)
        resp = self.session.post(url, files=files)
        if resp.status_code != 200:
            logger.error("Start pre-change analysis failed: {}".format(resp.json()))
            return resp, None

        try:
            job_id = json.loads(resp.content)["value"]["data"]["job_id"]
            logger.info("Pre-change analysis started. Job ID: {}".format(job_id))
            return None, job_id
        except KeyError:
            pass
        logger.error("Job ID could not be found: {}".format(resp.json()))
        return resp, None

    def wait_pcv(
        self, job_id: str
    ) -> Tuple[Optional[requests.Response], Optional[str]]:
        """Wair for pre-change validation to complete and return epoch job ID"""
        if not self.authenticated:
            err = self._login()
            if err is not None:
                return err, None

        status = None
        start_time = datetime.now()
        while True:
            url = "{}/config-services/prechange-analysis/{}".format(
                self.api_url, job_id
            )
            resp = self.session.get(url)
            if resp.status_code != 200:
                logger.error(
                    "Get pre-change analysis status failed: {}".format(resp.json())
                )
                return resp, None
            try:
                status = json.loads(resp.content)["value"]["data"]["analysis_status"]
                if status == "COMPLETED":
                    break
            except KeyError:
                logger.error("Status could not be found: {}".format(resp.json()))
            delta_minutes = (datetime.now() - start_time).total_seconds() / 60
            if delta_minutes > self.timeout:
                break
            logger.info("Waiting for pre-change analysis to complete ...")
            time.sleep(10)

        try:
            epoch_job_id = json.loads(resp.content)["value"]["data"][
                "epoch_delta_job_id"
            ]
            logger.info(
                "Pre-change analysis completed. Epoch job ID: {}".format(epoch_job_id)
            )
            return None, epoch_job_id
        except KeyError:
            pass
        logger.error("Epoch job ID could not be found: {}".format(resp.json()))
        return resp, None

    def get_pcv_results(
        self, epoch_job_id: str, suppress_events: str
    ) -> Tuple[Optional[requests.Response], Optional[List[Any]]]:
        """Retrieve pre-change validation results"""
        if not self.authenticated:
            err = self._login()
            if err is not None:
                return err, None

        suppress_events_list = suppress_events.split(",")

        url = "{}/epoch-delta-services/assured-networks/{}/job/{}/health/view/aggregate-table?category=ADC,CHANGE_ANALYSIS,TENANT_ENDPOINT,TENANT_FORWARDING,TENANT_SECURITY,RESOURCE_UTILIZATION,SYSTEM,COMPLIANCE&epoch_status=EPOCH2_ONLY&severity=EVENT_SEVERITY_CRITICAL,EVENT_SEVERITY_MAJOR,EVENT_SEVERITY_MINOR,EVENT_SEVERITY_WARNING,EVENT_SEVERITY_INFO".format(
            self.api_url, self.assurance_group_uuid, epoch_job_id
        )
        resp = self.session.get(url)
        if resp.status_code != 200:
            logger.error("Get PCV results failed: {}".format(resp.json()))
            return resp, None

        event_list = []
        try:
            for event in json.loads(resp.content)["value"]["data"]:
                if int(event["count"]) > 0:
                    if (
                        str(event["epoch2_details"]["severity"])
                        == "EVENT_SEVERITY_INFO"
                        or str(event["epoch2_details"]["mnemonic"])
                        in suppress_events_list
                    ):
                        continue
                    event_list.append(
                        {
                            "Category": event["category"].title(),
                            "Count": event["count"],
                            "Description": event["epoch2_details"]["description"],
                            "Severity": event["epoch2_details"]["severity"][
                                15:
                            ].title(),
                        }
                    )
        except KeyError:
            logger.error("Could not find events: {}".format(resp.json()))
            return resp, None
        if event_list:
            logger.error(
                "The following events have been raised:\n{}".format(
                    yaml.dump(event_list)
                )
            )
        return None, event_list

    def get_pcv_url(
        self, job_id: str
    ) -> Tuple[Optional[requests.Response], Optional[str]]:
        """Get URL pointing to pre-change validation results"""
        if not self.authenticated:
            err = self._login()
            if err is not None:
                return err, None

        url = "{}/config-services/prechange-analysis/{}".format(self.api_url, job_id)
        resp = self.session.get(url)
        if resp.status_code != 200:
            logger.error("Get PCV url failed: {}".format(resp.json()))
            return resp, None
        try:
            response = json.loads(resp.content)["value"]["data"]
            pcv_url = "https://{}/nae/epoch-delta-analysis/result/health?analysis_id={}&type=PCV_EPOCH_DELTA_ANALYSIS&fabric_id={}&epoch={}#top".format(
                self.hostname_ip,
                response["epoch_delta_job_id"],
                response["fabric_uuid"],
                response["pre_change_epoch_uuid"],
            )
            return None, pcv_url
        except KeyError:
            pass
        logger.error("PCV url parameters could not be found: {}".format(resp.json()))
        return resp, None
